def test_get_config_yaml_item():
    from dcspy import get_config_yaml_item

    assert get_config_yaml_item('f16_ded_font')
    assert get_config_yaml_item('f16_ded_font', False)
    assert get_config_yaml_item('unknown', True)



from unittest.mock import patch

from pytest import mark


@mark.parametrize('function, args, result', [
    ('logi_led_init', (), False),
    ('logi_led_init_with_name', ('name',), False),
    ('logi_led_set_target_device', (1,), False),
    ('logi_led_save_current_lighting', (), False),
    ('logi_led_restore_lighting', (), False),
    ('logi_led_set_lighting', ((1, 2, 3),), False),
    ('logi_led_flash_lighting', ((1, 2, 3), 1, 1), False),
    ('logi_led_pulse_lighting', ((1, 2, 3), 1, 1), False),
    ('logi_led_stop_effects', (), False),
    ('logi_led_shutdown', (), None)
], ids=[
    'init',
    'init with name',
    'set target device',
    'save current lighting',
    'restore lighting',
    'set lighting',
    'flash lighting',
    'pulse lighting',
    'stop effects',
    'shutdown',
])
def test_all_failure_cases(function, args, result):
    from dcspy.sdk import led_sdk
    led_sdk.LED_DLL = None
    assert getattr(led_sdk, function)(*args) is result


@mark.slow
def test_start_led_pulse():
    from concurrent.futures import ThreadPoolExecutor
    from threading import Event
    from time import sleep

    from dcspy.sdk import led_sdk

    rgb = (100, 0, 0)
    duration = 1
    interval = 1
    event = Event()

    def pulse_led(_rgb, _duration, _interval, _event):
        with patch.object(led_sdk, 'logi_led_init', return_value=True) as logi_led_init, \
                patch.object(led_sdk, 'logi_led_set_target_device', return_value=True) as logi_led_set_target_device, \
                patch.object(led_sdk, 'logi_led_pulse_lighting', return_value=True) as logi_led_pulse_lighting, \
                patch.object(led_sdk, 'logi_led_shutdown', return_value=True) as logi_led_shutdown:
            led_sdk.start_led_pulse(_rgb, _duration, _interval, _event)
            logi_led_init.assert_called_once()
            logi_led_set_target_device.assert_called_once_with(3)
            logi_led_pulse_lighting.assert_called_once_with(_rgb, _duration, _interval)
            logi_led_shutdown.assert_called_once()
        return True

    with ThreadPoolExecutor() as executor:
        future = executor.submit(pulse_led, rgb, duration, interval, event)
        sleep(1.3)
        event.set()
        assert future.result()



from unittest.mock import call, patch

from pytest import mark


@mark.parametrize('function, args, result', [
    ('logi_lcd_init', ('test', 1), False),
    ('logi_lcd_is_connected', (1,), False),
    ('logi_lcd_is_button_pressed', (1,), False),
    ('logi_lcd_update', (), None),
    ('logi_lcd_shutdown', (), None),
    ('logi_lcd_mono_set_background', ([1, 2, 3],), False),
    ('logi_lcd_mono_set_text', (1, ''), False),
    ('logi_lcd_color_set_background', ([(1, 2, 3)],), False),
    ('logi_lcd_color_set_title', ('', (1, 2, 3)), False),
    ('logi_lcd_color_set_text', (1, '', (1, 2, 3)), False)
], ids=[
    'init',
    'is connected',
    'is button pressed',
    'update',
    'shutdown',
    'mono set background',
    'mono set text',
    'color set background',
    'color_set title',
    'color set text',
])
def test_all_failure_cases(function, args, result):
    from dcspy.sdk import lcd_sdk
    lcd_sdk.LCD_DLL = None
    assert getattr(lcd_sdk, function)(*args) is result


@mark.parametrize('c_func, effect, lcd, size', [
    ('logi_lcd_mono_set_background', [True], 1, (16, 4)),
    ('logi_lcd_color_set_background', [False, True], 2, (32, 24))
], ids=['Mono', 'Color'])
def test_update_display(c_func, effect, lcd, size):
    from PIL import Image

    from dcspy.sdk import lcd_sdk
    with patch.object(lcd_sdk, 'logi_lcd_is_connected', side_effect=effect) as connected, \
            patch.object(lcd_sdk, c_func, return_value=True) as set_background, \
            patch.object(lcd_sdk, 'logi_lcd_update', return_value=True):
        lcd_sdk.update_display(Image.new('1', (size[0], size[1]), 0))
        connected.assert_called_with(lcd)
        set_background.assert_called_once_with([0] * size[0] * size[1])


@mark.parametrize('c_func, effect, lcd, list_txt', [
    ('logi_lcd_mono_set_text', [True], 1, ['1', '2', '3', '4']),
    ('logi_lcd_color_set_text', [False, True], 2, ['1', '2', '3', '4', '5', '6', '7', '8'])
], ids=['Mono', 'Color'])
def test_update_text(c_func, effect, lcd, list_txt):
    from dcspy.sdk import lcd_sdk
    with patch.object(lcd_sdk, 'logi_lcd_is_connected', side_effect=effect) as connected:
        with patch.object(lcd_sdk, c_func, return_value=True) as set_text:
            with patch.object(lcd_sdk, 'logi_lcd_update', return_value=True):
                lcd_sdk.update_text(list_txt)
                connected.assert_called_with(lcd)
                set_text.assert_has_calls([call(i, j) for i, j in enumerate(list_txt)])


@mark.parametrize('c_funcs, effect, lcd, clear, text', [
    (('logi_lcd_mono_set_background', 'logi_lcd_mono_set_text'), [True], 1, [0] * 160 * 43, [call(0, ''), call(1, ''), call(2, ''), call(3, '')]),
    (('logi_lcd_color_set_background', 'logi_lcd_color_set_text'), [False, True], 2, [(0,) * 4] * 320 * 240,
     [call(0, ''), call(1, ''), call(2, ''), call(3, ''), call(4, ''), call(5, ''), call(6, ''), call(7, '')])
], ids=['Mono', 'Color'])
def test_clear_display(c_funcs, effect, lcd, clear, text):
    from dcspy.sdk import lcd_sdk
    with patch.object(lcd_sdk, 'logi_lcd_is_connected', side_effect=effect) as connected, \
            patch.object(lcd_sdk, c_funcs[0], return_value=True) as set_background, \
            patch.object(lcd_sdk, c_funcs[1], return_value=True) as set_text, \
            patch.object(lcd_sdk, 'logi_lcd_update', return_value=True):
        lcd_sdk.clear_display(true_clear=True)
        connected.assert_called_with(lcd)
        set_background.assert_called_once_with(clear)
        set_text.assert_has_calls(text)


def test_update_text_no_lcd():
    from dcspy.sdk import lcd_sdk
    with patch.object(lcd_sdk, 'logi_lcd_is_connected', side_effect=[False, False]) as connected:
        lcd_sdk.update_text(['1'])
        connected.assert_has_calls([call(1), call(2)])


def test_update_display_no_lcd():
    from PIL import Image

    from dcspy.sdk import lcd_sdk
    with patch.object(lcd_sdk, 'logi_lcd_is_connected', side_effect=[False, False]) as connected:
        lcd_sdk.update_display(Image.new('1', (16, 4), 0))
        connected.assert_has_calls([call(1), call(2)])



from logging import getLogger
from os import name
from pathlib import Path
from platform import architecture, python_implementation, python_version, uname
from sys import executable, platform
from typing import Optional, Union

from dcspy.log import config_logger
from dcspy.migration import migrate
from dcspy.models import LOCAL_APPDATA
from dcspy.utils import check_dcs_ver, get_default_yaml, load_yaml, save_yaml

__version__ = '3.0.0'

default_yaml = get_default_yaml(local_appdata=LOCAL_APPDATA)
_config = migrate(load_yaml(full_path=default_yaml))
save_yaml(data=_config, full_path=default_yaml)
LOG = getLogger(__name__)
config_logger(LOG, _config['verbose'])

LOG.debug(f'Arch: {name} / {platform} / {" / ".join(architecture())}')
LOG.debug(f'Python: {python_implementation()}-{python_version()}')
LOG.debug(f'Python exec: {executable}')
LOG.debug(f'{uname()}')
LOG.debug(f'Configuration: {_config} from: {default_yaml}')
LOG.info(f'dcspy {__version__} https://github.com/emcek/dcspy')
dcs_type, dcs_ver = check_dcs_ver(Path(str(_config['dcs'])))
LOG.info(f'DCS {dcs_type} ver: {dcs_ver}')


def get_config_yaml_item(key: str, /, default: Optional[Union[str, int]] = None) -> Union[str, int]:
    """
    Get item from configuration YAML file.

    :param key: key to get
    :param default: default value if key not found
    :return: value from configuration
    """
    return load_yaml(full_path=default_yaml).get(key, default)



import json
import sys
import zipfile
from datetime import datetime
from glob import glob
from itertools import chain
from logging import getLogger
from os import environ, makedirs, walk
from pathlib import Path
from platform import python_implementation, python_version, uname
from pprint import pformat
from re import search
from shutil import rmtree
from subprocess import CalledProcessError, run
from tempfile import gettempdir
from typing import Any, Dict, List, Optional, Tuple, Union

import yaml
from packaging import version
from psutil import process_iter
from requests import get

from dcspy.models import CTRL_LIST_SEPARATOR, ControlKeyData, DcsBiosPlaneData, DcspyConfigYaml, ReleaseInfo

try:
    import git
except ImportError:
    pass

LOG = getLogger(__name__)
__version__ = '3.0.0'
CONFIG_YAML = 'config.yaml'
DEFAULT_YAML_FILE = Path(__file__).resolve().with_name(CONFIG_YAML)

with open(DEFAULT_YAML_FILE) as c_file:
    defaults_cfg: DcspyConfigYaml = yaml.load(c_file, Loader=yaml.FullLoader)
    defaults_cfg['dcsbios'] = f'D:\\Users\\{environ.get("USERNAME", "UNKNOWN")}\\Saved Games\\DCS.openbeta\\Scripts\\DCS-BIOS'


def get_default_yaml(local_appdata=False) -> Path:
    """
    Return full path to default configuration file.

    :param local_appdata: if True C:/Users/<user_name>/AppData/Local is used
    :return: Path like object
    """
    cfg_ful_path = DEFAULT_YAML_FILE
    if local_appdata:
        localappdata = environ.get('LOCALAPPDATA', None)
        user_appdata = Path(localappdata) / 'dcspy' if localappdata else DEFAULT_YAML_FILE.parent
        makedirs(name=user_appdata, exist_ok=True)
        cfg_ful_path = Path(user_appdata / CONFIG_YAML).resolve()
        if not cfg_ful_path.exists():
            save_yaml(data=defaults_cfg, full_path=cfg_ful_path)
    return cfg_ful_path


def load_yaml(full_path: Path) -> Dict[str, Any]:
    """
    Load yaml from file into dictionary.

    :param full_path: full path to yaml file
    :return: dictionary
    """
    try:
        with open(file=full_path, encoding='utf-8') as yaml_file:
            data = yaml.load(yaml_file, Loader=yaml.FullLoader)
            if not isinstance(data, dict):
                data = {}
    except (FileNotFoundError, yaml.parser.ParserError) as err:
        makedirs(name=full_path.parent, exist_ok=True)
        LOG.warning(f'{type(err).__name__}: {full_path}.')
        LOG.debug(f'{err}')
        data = {}
    return data


def save_yaml(data: Dict[str, Any], full_path: Path) -> None:
    """
    Save disc as yaml file.

    :param data: dict
    :param full_path: full path to yaml file
    """
    with open(file=full_path, mode='w', encoding='utf-8') as yaml_file:
        yaml.dump(data, yaml_file)


def check_ver_at_github(repo: str, current_ver: str, extension: str) -> ReleaseInfo:
    """
    Check version of <organization>/<package> at GitHub.

    Return tuple with:
    - result (bool) - if local version is latest
    - online version (version.Version) - the latest version
    - download url (str) - ready to download
    - published date (str) - format DD MMMM YYYY
    - release type (str) - Regular or Pre-release
    - asset file (str) - file name of asset

    :param repo: format '<organization or user>/<package>'
    :param current_ver: current local version
    :param extension: file extension to be returned
    :return: ReleaseInfo NamedTuple with information
    """
    latest, online_version, asset_url, published, pre_release = False, '0.0.0', '', '', False
    package = repo.split('/')[1]
    try:
        response = get(url=f'https://api.github.com/repos/{repo}/releases/latest', timeout=5)
        if response.ok:
            dict_json = response.json()
            online_version = dict_json['tag_name']
            pre_release = dict_json['prerelease']
            published = datetime.strptime(dict_json['published_at'], '%Y-%m-%dT%H:%M:%S%z').strftime('%d %B %Y')
            asset_url = next(url for url in [asset['browser_download_url'] for asset in dict_json['assets']] if url.endswith(extension))
            LOG.debug(f'Latest GitHub version:{online_version} pre:{pre_release} date:{published} url:{asset_url}')
            latest = _compare_versions(package, current_ver, online_version)
        else:
            LOG.warning(f'Unable to check {package} version online. Try again later. Status={response.status_code}')
    except Exception as exc:
        LOG.warning(f'Unable to check {package} version online: {exc}')
    return ReleaseInfo(
        latest=latest,
        ver=version.parse(online_version),
        dl_url=asset_url,
        published=published,
        release_type='Pre-release' if pre_release else 'Regular',
        asset_file=asset_url.split('/')[-1],
    )


def _compare_versions(package: str, current_ver: str, remote_ver: str) -> bool:
    """
    Compare two version of package and return result.

    :param package: package name
    :param current_ver: current/local version
    :param remote_ver: remote/online version
    :return:
    """
    latest = False
    if version.parse(remote_ver) > version.parse(current_ver):
        LOG.info(f'There is new version of {package}: {remote_ver}')
    elif version.parse(remote_ver) <= version.parse(current_ver):
        LOG.info(f'{package} is up-to-date version: {current_ver}')
        latest = True
    return latest


def get_version_string(repo: str, current_ver: str, check=True) -> str:
    """
    Generate formatted string with version number.

    :param repo: format '<organization or user>/<package>'.
    :param current_ver: current local version.
    :param check: version online.
    :return: formatted version as string.
    """
    ver_string = f'v{current_ver}'
    if check:
        result = check_ver_at_github(repo=repo, current_ver=current_ver, extension='')
        details = ''
        if result.latest:
            details = ' (latest)'
        elif str(result.ver) != '0.0.0':
            details = ' (update!)'
        elif str(result.ver) == '0.0.0':
            details = ' (failed)'
        ver_string = f'v{current_ver}{details}'
    return ver_string


def download_file(url: str, save_path: Path) -> bool:
    """
    Download file from URL and save to save_path.

    :param url: URL address
    :param save_path: full path to save
    """
    response = get(url=url, stream=True, timeout=5)
    if response.ok:
        LOG.debug(f'Download file from: {url}')
        with open(save_path, 'wb+') as dl_file:
            for chunk in response.iter_content(chunk_size=128):
                dl_file.write(chunk)
            LOG.debug(f'Saved as: {save_path}')
            return True
    else:
        LOG.warning(f'Can not download from: {url}')
        return False


def proc_is_running(name: str) -> int:
    """
    Check if process is running and return its PID.

    If process name is not found, 0 (zero) is returned.
    :param name: process name
    :return: PID as int
    """
    for proc in process_iter(['pid', 'name']):
        if name in proc.info['name']:
            return proc.info['pid']
    return 0


def check_dcs_ver(dcs_path: Path) -> Tuple[str, str]:
    """
    Check DCS version and release type.

    :param dcs_path: path to DCS installation directory
    :return: dcs type and version as strings
    """
    result_type, result_ver = 'Unknown', 'Unknown'
    try:
        with open(file=dcs_path / 'autoupdate.cfg', encoding='utf-8') as autoupdate_cfg:
            autoupdate_data = autoupdate_cfg.read()
    except (FileNotFoundError, PermissionError) as err:
        LOG.debug(f'{type(err).__name__}: {err.filename}')
    else:
        result_type = 'stable'
        dcs_type = search(r'"branch":\s"([\w.]*)"', autoupdate_data)
        if dcs_type:
            result_type = str(dcs_type.group(1))
        dcs_ver = search(r'"version":\s"([\d.]*)"', autoupdate_data)
        if dcs_ver:
            result_ver = str(dcs_ver.group(1))
    return result_type, result_ver


def check_bios_ver(bios_path: Union[Path, str]) -> ReleaseInfo:
    """
    Check DSC-BIOS release version.

    :param bios_path: path to DCS-BIOS directory in Saved Games folder
    :return: ReleaseInfo named tuple
    """
    result = ReleaseInfo(latest=False, ver=version.parse('0.0.0'), dl_url='', published='', release_type='', asset_file='')
    try:
        with open(file=Path(bios_path) / 'lib' / 'modules' / 'common_modules' / 'CommonData.lua', encoding='utf-8') as cd_lua:
            cd_lua_data = cd_lua.read()
    except FileNotFoundError as err:
        LOG.debug(f'While checking DCS-BIOS version {type(err).__name__}: {err.filename}')
    else:
        bios_re = search(r'function getVersion\(\)\s*return\s*\"([\d.]*)\"', cd_lua_data)
        if bios_re:
            bios = version.parse(bios_re.group(1))
            result = ReleaseInfo(latest=False, ver=bios, dl_url='', published='', release_type='', asset_file='')
    return result


def is_git_repo(dir_path: str) -> bool:
    """
    Check if dir_path ios Git repository.

    :param dir_path: path as string
    :return: true if dir is git repo
    """
    import git
    try:
        _ = git.Repo(dir_path).git_dir
        return True
    except (git.InvalidGitRepositoryError, git.exc.NoSuchPathError):
        return False


def check_github_repo(git_ref: str, update=True, repo='DCS-Skunkworks/dcs-bios', repo_dir=Path(gettempdir()) / 'dcsbios_git') -> str:
    """
    Update DCS-BIOS git repository.

    Return SHA of latest commit.

    :param git_ref: any Git reference as string
    :param update: perform update process
    :param repo: GitHub repository
    :param repo_dir: local directory for repository
    """
    try:
        import git
    except ImportError:
        raise OSError('Git executable is not available!')

    bios_repo = _checkout_repo(repo, repo_dir)
    if update:
        f_info = bios_repo.remotes[0].pull()
        LOG.debug(f'Pulled: {f_info[0].name} as: {f_info[0].commit}')
        try:
            bios_repo.git.checkout(git_ref)
            branch = bios_repo.active_branch.name
            head_commit = bios_repo.head.commit
            sha = f'{branch}: {head_commit.committed_datetime} by: {head_commit.author}'
        except (git.exc.GitCommandError, TypeError):
            head_commit = bios_repo.head.commit
            sha = f'{head_commit.hexsha[0:8]} from: {head_commit.committed_datetime} by: {head_commit.author}'
        LOG.debug(f'Checkout: {head_commit.hexsha} from: {head_commit.committed_datetime} | by: {head_commit.author}\n{head_commit.message}')  # type: ignore
    else:
        bios_repo.git.checkout(git_ref)
        head_commit = bios_repo.head.commit
        sha = f'{head_commit.hexsha[0:8]} from: {head_commit.committed_datetime.strftime("%d-%b-%Y %H:%M:%S")}'
    return sha


def _checkout_repo(repo: str, repo_dir: Path, checkout_ref: str = 'master') -> 'git.Repo':
    """
    Checkout repository at master branch or clone it when not exists in system.

    :param repo: repository name
    :param repo_dir: local repository directory
    :param checkout_ref: git reference to checkout
    :return: Repo object to repository
    """
    import git

    makedirs(name=repo_dir, exist_ok=True)
    if is_git_repo(str(repo_dir)):
        bios_repo = git.Repo(repo_dir)
        bios_repo.git.checkout(checkout_ref)
    else:
        rmtree(path=repo_dir, ignore_errors=True)
        bios_repo = git.Repo.clone_from(url=f'https://github.com/{repo}.git', to_path=repo_dir)
    return bios_repo


def check_dcs_bios_entry(lua_dst_data: str, lua_dst_path: Path, temp_dir: Path) -> str:
    """
    Check DCS-BIOS entry in Export.lua file.

    :param lua_dst_data: content of Export.lua
    :param lua_dst_path: Export.lua path
    :param temp_dir: directory with DCS-BIOS archive
    :return: result of checks
    """
    result = '\n\nExport.lua exists.'
    lua = 'Export.lua'
    with open(file=temp_dir / lua, encoding='utf-8') as lua_src:
        lua_src_data = lua_src.read()
    export_re = search(r'dofile\(lfs.writedir\(\)\.\.\[\[Scripts\\DCS-BIOS\\BIOS\.lua\]\]\)', lua_dst_data)
    if not export_re:
        with open(file=lua_dst_path / lua, mode='a+', encoding='utf-8') as exportlua_dst:
            exportlua_dst.write(f'\n{lua_src_data}')
        LOG.debug(f'Add DCS-BIOS to Export.lua: {lua_src_data}')
        result += '\n\nDCS-BIOS entry added.\n\nYou verify installation at:\ngithub.com/DCS-Skunkworks/DCSFlightpanels/wiki/Installation'
    else:
        result += '\n\nDCS-BIOS entry detected.'
    return result


def is_git_exec_present() -> bool:
    """
    Check if git executable is present in system.

    :return: True if git.exe is available
    """
    try:
        import git
        return bool(git.GIT_OK)
    except ImportError as err:
        LOG.debug(type(err).__name__, exc_info=True)
        return False


def is_git_object(repo_dir: Path, git_obj: str) -> bool:
    """
    Check if git_obj is valid Git reference.

    :param repo_dir: directory with repository
    :param git_obj: git reference to check
    :return: True if git_obj is git reference, False otherwise
    """
    import gitdb
    result = False
    if is_git_repo(str(repo_dir)):
        bios_repo = git.Repo(repo_dir)
        bios_repo.git.checkout('master')
        try:
            bios_repo.commit(git_obj)
            result = True
        except gitdb.exc.BadName:
            pass
    return result


def get_all_git_refs(repo_dir: Path) -> List[str]:
    """
    Get list of branches and tags for repo.

    :param repo_dir: directory with repository
    :return: list of git references as  strings
    """
    refs = []
    if is_git_repo(str(repo_dir)):
        for ref in chain(git.Repo(repo_dir).heads, git.Repo(repo_dir).tags):
            refs.append(str(ref))
    return refs


def get_sha_for_current_git_ref(git_ref: str, repo='DCS-Skunkworks/dcs-bios', repo_dir=Path(gettempdir()) / 'dcsbios_git') -> str:
    """
    Get SHA for current git reference.

    :param git_ref: any Git reference as string
    :param repo: GitHub repository
    :param repo_dir: local directory for repository
    :return: Hex of SHA
    """
    bios_repo = _checkout_repo(repo=repo, repo_dir=repo_dir, checkout_ref=git_ref)
    head_commit = bios_repo.head.commit
    return head_commit.hexsha


def collect_debug_data() -> Path:
    """
    Collect add zipp all data for troubleshooting.

    :return: Path object to zip file
    """
    aircrafts = ['FA18Chornet', 'Ka50', 'Ka503', 'Mi8MT', 'Mi24P', 'F16C50', 'F15ESE', 'AH64DBLKII', 'A10C', 'A10C2', 'F14A135GR', 'F14B', 'AV8BNA']
    localappdata = environ.get('LOCALAPPDATA', None)
    user_appdata = Path(localappdata) / 'dcspy' if localappdata else DEFAULT_YAML_FILE.parent
    config_file = Path(user_appdata / CONFIG_YAML).resolve()

    conf_dict = load_yaml(config_file)
    name = uname()
    pyver = (python_version(), python_implementation())
    pyexec = sys.executable
    dcs = check_dcs_ver(dcs_path=Path(str(conf_dict['dcs'])))
    bios_ver = check_bios_ver(bios_path=str(conf_dict['dcsbios'])).ver
    git_ver = (0, 0, 0, 0)
    head_commit = 'N/A'
    try:
        import git
        git_ver = git.cmd.Git().version_info
        head_commit = git.Repo(Path(gettempdir()) / 'dcsbios_git').head.commit
    except (git.exc.NoSuchPathError, ImportError):
        pass

    lgs_dir = '\n'.join([
        str(Path(dirpath) / filename)
        for dirpath, _, filenames in walk('C:\\Program Files\\Logitech Gaming Software\\SDK')
        for filename in filenames
    ])

    png_files = [
        Path(dirpath) / filename
        for dirpath, _, filenames in walk(gettempdir())
        for filename in filenames
        if any(True for aircraft in aircrafts if aircraft in filename and filename.endswith('png'))
    ]

    log_files = []
    for logfile in glob(str(Path(gettempdir()) / 'dcspy.log*')):
        log_files.append(Path(Path(gettempdir()) / logfile))
    sys_data = Path(gettempdir()) / 'system_data.txt'
    zip_file = Path(gettempdir()) / f'dcspy_debug_{str(datetime.now()).replace(" ", "_").replace(":", "")}.zip'

    with open(sys_data, 'w+') as debug_file:
        debug_file.write(f'{__version__=}\n{name=}\n{pyver=}\n{pyexec=}\n{dcs=}\n{bios_ver=}\n{git_ver=}\n{head_commit=}\n{lgs_dir}\ncfg={pformat(conf_dict)}')

    with zipfile.ZipFile(file=zip_file, mode='w', compresslevel=9, compression=zipfile.ZIP_DEFLATED) as zipf:
        zipf.write(sys_data, arcname=sys_data.name)
        for log_file in log_files:
            zipf.write(log_file, arcname=log_file.name)
        zipf.write(config_file, arcname=config_file.name)
        for png in png_files:
            zipf.write(png, arcname=png.name)

    return zip_file


def run_pip_command(cmd: str) -> Tuple[int, str, str]:
    """
    Execute pip command.

    :param cmd: as string
    :return: tuple with return code, stderr and stdout
    """
    try:
        result = run([sys.executable, '-m', 'pip', *cmd.split(' ')], capture_output=True, check=True)
        return result.returncode, result.stderr.decode('utf-8'), result.stdout.decode('utf-8')
    except CalledProcessError as e:
        LOG.debug(f'Result: {e}')
        return e.returncode, e.stderr.decode('utf-8'), e.stdout.decode('utf-8')


def load_json(path: Path) -> Dict[str, Any]:
    """
    Load json from file into dictionary.

    :param path: full path
    :return: dict
    """
    with open(path, encoding='utf-8') as json_file:
        data = json_file.read()
    return json.loads(data)


def get_full_bios_for_plane(plane: str, bios_dir: Path) -> DcsBiosPlaneData:
    """
    Collect full BIOS for plane with name.

    :param plane: BIOS plane name
    :param bios_dir: path to DCS-BIOS directory
    :return: dict
    """
    alias_path = bios_dir / 'doc' / 'json' / 'AircraftAliases.json'
    local_json: Dict[str, Any] = {}
    aircraft_aliases = load_json(path=alias_path)
    for json_file in aircraft_aliases[plane]:
        local_json = {**local_json, **load_json(path=bios_dir / 'doc' / 'json' / f'{json_file}.json')}

    return DcsBiosPlaneData.model_validate(local_json)


def get_inputs_for_plane(plane: str, bios_dir: Path) -> Dict[str, Dict[str, ControlKeyData]]:
    """
    Get dict with all not empty inputs for plane.

    :param plane: BIOS plane name
    :param bios_dir: path to DCS-BIOS
    :return: dict.
    """
    plane_bios = get_full_bios_for_plane(plane=plane, bios_dir=bios_dir)
    inputs = plane_bios.get_inputs()
    return inputs


def get_list_of_ctrls(inputs: Dict[str, Dict[str, ControlKeyData]]) -> List[str]:
    """
    Get list of all controllers from dict with sections and inputs.

    :param inputs: dict with ControlKeyData
    :return: list of string
    """
    result_list = []
    for section, controllers in inputs.items():
        result_list.append(f'{CTRL_LIST_SEPARATOR} {section} {CTRL_LIST_SEPARATOR}')
        for ctrl_name in controllers:
            result_list.append(ctrl_name)
    return result_list


def get_planes_list(bios_dir: Path) -> List[str]:
    """
    Get list of all DCS-BIOS supported planes with clickable cockpit.

    :param bios_dir: path to DCS-BIOS
    :return: list of all supported planes
    """
    aircraft_aliases = get_plane_aliases(bios_dir=bios_dir, plane=None)
    return [name for name, yaml_data in aircraft_aliases.items() if yaml_data not in (['CommonData', 'FC3'], ['CommonData'])]


def get_plane_aliases(bios_dir: Path, plane: Optional[str] = None) -> Dict[str, List[str]]:
    """
    Get list of all yaml files for plane with name.

    :param plane: BIOS plane name
    :param bios_dir: path to DCS-BIOS
    :return: list of all yaml files for plane definition
    """
    alias_path = bios_dir / 'doc' / 'json' / 'AircraftAliases.json'
    aircraft_aliases = load_json(path=alias_path)
    if plane:
        aircraft_aliases = {plane: aircraft_aliases[plane]}
    return aircraft_aliases



from pathlib import Path
from sys import platform
from unittest.mock import patch

from pytest import mark, raises

from dcspy.models import LcdButton, ZigZagIterator
from tests.helpers import all_plane_list, compare_images, set_bios_during_test


# <=><=><=><=><=> Base Class <=><=><=><=><=>
@mark.parametrize('plane', all_plane_list)
def test_check_all_aircraft_inherit_from_correct_base_class(plane, request):
    from dcspy import aircraft
    plane = request.getfixturevalue(f'{plane}_mono')
    assert isinstance(plane, aircraft.BasicAircraft)


@mark.parametrize('selector, data, value, c_func, effect, plane', [
    ('field1', {'addr': 0xdeadbeef, 'len': 16, 'value': ''}, 'val1', 'logi_lcd_mono_set_background', [True], 'advancedaircraft_mono'),
    ('field2', {'addr': 0xdeadbeef, 'len': 16, 'value': ''}, 'val2', 'logi_lcd_color_set_background', [False, True], 'advancedaircraft_color'),
], ids=['Mono LCD', 'Color LCD'])
def test_aircraft_base_class_set_bios(selector, data, value, c_func, effect, plane, request):
    from dcspy.sdk import lcd_sdk
    aircraft = request.getfixturevalue(plane)
    assert aircraft.bios_data == {}
    aircraft.bios_data = {selector: data}
    with patch.object(lcd_sdk, 'logi_lcd_is_connected', side_effect=effect), \
            patch.object(lcd_sdk, c_func, return_value=True), \
            patch.object(lcd_sdk, 'logi_lcd_update', return_value=True):
        assert aircraft.bios_data[selector]['value'] == ''
        assert aircraft.get_bios('none') == ''
        with raises(NotImplementedError):
            aircraft.set_bios(selector, value)


@mark.parametrize('c_func, plane', [
    ('logi_lcd_mono_set_background', 'advancedaircraft_mono'),
    ('logi_lcd_color_set_background', 'advancedaircraft_color'),
], ids=['Mono LCD', 'Color LCD'])
def test_aircraft_base_class_prepare_img(c_func, plane, request):
    from dcspy.sdk import lcd_sdk
    aircraft = request.getfixturevalue(plane)
    with patch.object(lcd_sdk, 'logi_lcd_is_connected', return_value=True), \
            patch.object(lcd_sdk, c_func, return_value=True), \
            patch.object(lcd_sdk, 'logi_lcd_update', return_value=True), \
            raises(NotImplementedError):
        aircraft.prepare_image()


@mark.parametrize('keyboard, plane_name', [
    ('keyboard_mono', 'F-22A'),
    ('keyboard_color', 'UH-60L'),
], ids=['F-22A Mono Keyboard', 'UH-60L Color Keyboard'])
def test_meta_plane(keyboard, plane_name, request):
    from dcspy.aircraft import BasicAircraft, MetaAircraft

    keyboard = request.getfixturevalue(keyboard)
    plane = MetaAircraft(plane_name, (BasicAircraft,), {})(keyboard.lcd)
    assert isinstance(plane, BasicAircraft)
    assert type(plane).__name__ == plane_name


# <=><=><=><=><=> Button Requests <=><=><=><=><=>
@mark.parametrize('plane, button, result', [
    ('fa18chornet_mono', LcdButton.NONE, '\n'),
    ('fa18chornet_mono', LcdButton.ONE, 'UFC_COMM1_CHANNEL_SELECT DEC\n'),
    ('fa18chornet_mono', LcdButton.TWO, 'UFC_COMM1_CHANNEL_SELECT INC\n'),
    ('fa18chornet_mono', LcdButton.THREE, 'UFC_COMM2_CHANNEL_SELECT DEC\n'),
    ('fa18chornet_mono', LcdButton.FOUR, 'UFC_COMM2_CHANNEL_SELECT INC\n'),
    ('fa18chornet_color', LcdButton.NONE, '\n'),
    ('fa18chornet_color', LcdButton.LEFT, 'UFC_COMM1_CHANNEL_SELECT DEC\n'),
    ('fa18chornet_color', LcdButton.RIGHT, 'UFC_COMM1_CHANNEL_SELECT INC\n'),
    ('fa18chornet_color', LcdButton.DOWN, 'UFC_COMM2_CHANNEL_SELECT DEC\n'),
    ('fa18chornet_color', LcdButton.UP, 'UFC_COMM2_CHANNEL_SELECT INC\n'),
    ('fa18chornet_color', LcdButton.MENU, 'IFEI_DWN_BTN 1\n'),
    ('fa18chornet_color', LcdButton.CANCEL, 'IFEI_UP_BTN 1\n'),
    ('fa18chornet_color', LcdButton.OK, 'HUD_ATT_SW 1\n'),
    ('av8bna_mono', LcdButton.NONE, '\n'),
    ('av8bna_mono', LcdButton.ONE, 'UFC_COM1_SEL -3200\n'),
    ('av8bna_mono', LcdButton.TWO, 'UFC_COM1_SEL 3200\n'),
    ('av8bna_mono', LcdButton.THREE, 'UFC_COM2_SEL -3200\n'),
    ('av8bna_mono', LcdButton.FOUR, 'UFC_COM2_SEL 3200\n'),
    ('av8bna_color', LcdButton.NONE, '\n'),
    ('av8bna_color', LcdButton.LEFT, 'UFC_COM1_SEL -3200\n'),
    ('av8bna_color', LcdButton.RIGHT, 'UFC_COM1_SEL 3200\n'),
    ('av8bna_color', LcdButton.DOWN, 'UFC_COM2_SEL -3200\n'),
    ('av8bna_color', LcdButton.UP, 'UFC_COM2_SEL 3200\n'),
    ('av8bna_color', LcdButton.MENU, '\n'),
    ('av8bna_color', LcdButton.CANCEL, '\n'),
    ('av8bna_color', LcdButton.OK, '\n'),
    ('f15ese_mono', LcdButton.NONE, '\n'),
    ('f15ese_mono', LcdButton.ONE, 'F_UFC_PRE_CHAN_L_SEL -3200\n'),
    ('f15ese_mono', LcdButton.TWO, 'F_UFC_PRE_CHAN_L_SEL 3200\n'),
    ('f15ese_mono', LcdButton.THREE, 'F_UFC_PRE_CHAN_R_SEL -3200\n'),
    ('f15ese_mono', LcdButton.FOUR, 'F_UFC_PRE_CHAN_R_SEL 3200\n'),
    ('f15ese_color', LcdButton.NONE, '\n'),
    ('f15ese_color', LcdButton.LEFT, 'F_UFC_PRE_CHAN_L_SEL -3200\n'),
    ('f15ese_color', LcdButton.RIGHT, 'F_UFC_PRE_CHAN_L_SEL 3200\n'),
    ('f15ese_color', LcdButton.DOWN, 'F_UFC_PRE_CHAN_R_SEL -3200\n'),
    ('f15ese_color', LcdButton.UP, 'F_UFC_PRE_CHAN_R_SEL 3200\n'),
    ('f15ese_color', LcdButton.MENU, 'F_UFC_KEY_L_GUARD 1\n|F_UFC_KEY_L_GUARD 0\n'),
    ('f15ese_color', LcdButton.CANCEL, 'F_UFC_KEY_R_GUARD 1\n|F_UFC_KEY_R_GUARD 0\n'),
    ('f15ese_color', LcdButton.OK, '\n'),
    ('ka50_mono', LcdButton.NONE, '\n'),
    ('ka50_mono', LcdButton.ONE, 'PVI_WAYPOINTS_BTN 1\n|PVI_WAYPOINTS_BTN 0\n'),
    ('ka50_mono', LcdButton.TWO, 'PVI_FIXPOINTS_BTN 1\n|PVI_FIXPOINTS_BTN 0\n'),
    ('ka50_mono', LcdButton.THREE, 'PVI_AIRFIELDS_BTN 1\n|PVI_AIRFIELDS_BTN 0\n'),
    ('ka50_mono', LcdButton.FOUR, 'PVI_TARGETS_BTN 1\n|PVI_TARGETS_BTN 0\n'),
    ('ka50_color', LcdButton.NONE, '\n'),
    ('ka50_color', LcdButton.LEFT, 'PVI_WAYPOINTS_BTN 1\n|PVI_WAYPOINTS_BTN 0\n'),
    ('ka50_color', LcdButton.RIGHT, 'PVI_FIXPOINTS_BTN 1\n|PVI_FIXPOINTS_BTN 0\n'),
    ('ka50_color', LcdButton.DOWN, 'PVI_AIRFIELDS_BTN 1\n|PVI_AIRFIELDS_BTN 0\n'),
    ('ka50_color', LcdButton.UP, 'PVI_TARGETS_BTN 1\n|PVI_TARGETS_BTN 0\n'),
    ('ka50_color', LcdButton.MENU, '\n'),
    ('ka50_color', LcdButton.CANCEL, '\n'),
    ('ka50_color', LcdButton.OK, '\n'),
    ('ka503_mono', LcdButton.NONE, '\n'),
    ('ka503_mono', LcdButton.ONE, 'PVI_WAYPOINTS_BTN 1\n|PVI_WAYPOINTS_BTN 0\n'),
    ('ka503_mono', LcdButton.TWO, 'PVI_FIXPOINTS_BTN 1\n|PVI_FIXPOINTS_BTN 0\n'),
    ('ka503_mono', LcdButton.THREE, 'PVI_AIRFIELDS_BTN 1\n|PVI_AIRFIELDS_BTN 0\n'),
    ('ka503_mono', LcdButton.FOUR, 'PVI_TARGETS_BTN 1\n|PVI_TARGETS_BTN 0\n'),
    ('ka503_color', LcdButton.NONE, '\n'),
    ('ka503_color', LcdButton.LEFT, 'PVI_WAYPOINTS_BTN 1\n|PVI_WAYPOINTS_BTN 0\n'),
    ('ka503_color', LcdButton.RIGHT, 'PVI_FIXPOINTS_BTN 1\n|PVI_FIXPOINTS_BTN 0\n'),
    ('ka503_color', LcdButton.DOWN, 'PVI_AIRFIELDS_BTN 1\n|PVI_AIRFIELDS_BTN 0\n'),
    ('ka503_color', LcdButton.UP, 'PVI_TARGETS_BTN 1\n|PVI_TARGETS_BTN 0\n'),
    ('ka503_color', LcdButton.MENU, '\n'),
    ('ka503_color', LcdButton.CANCEL, '\n'),
    ('ka503_color', LcdButton.OK, '\n'),
    ('f14a135gr_mono', LcdButton.NONE, '\n'),
    ('f14a135gr_mono', LcdButton.ONE, 'RIO_CAP_CLEAR 1\n|RIO_CAP_CLEAR 0\n'),
    ('f14a135gr_mono', LcdButton.TWO, 'RIO_CAP_SW 1\n|RIO_CAP_SW 0\n'),
    ('f14a135gr_mono', LcdButton.THREE, 'RIO_CAP_NE 1\n|RIO_CAP_NE 0\n'),
    ('f14a135gr_mono', LcdButton.FOUR, 'RIO_CAP_ENTER 1\n|RIO_CAP_ENTER 0\n'),
    ('f14a135gr_color', LcdButton.NONE, '\n'),
    ('f14a135gr_color', LcdButton.LEFT, 'RIO_CAP_CLEAR 1\n|RIO_CAP_CLEAR 0\n'),
    ('f14a135gr_color', LcdButton.RIGHT, 'RIO_CAP_SW 1\n|RIO_CAP_SW 0\n'),
    ('f14a135gr_color', LcdButton.DOWN, 'RIO_CAP_NE 1\n|RIO_CAP_NE 0\n'),
    ('f14a135gr_color', LcdButton.UP, 'RIO_CAP_ENTER 1\n|RIO_CAP_ENTER 0\n'),
    ('f14a135gr_color', LcdButton.MENU, '\n'),
    ('f14a135gr_color', LcdButton.CANCEL, '\n'),
    ('f14a135gr_color', LcdButton.OK, '\n'),
    ('f14b_mono', LcdButton.NONE, '\n'),
    ('f14b_mono', LcdButton.ONE, 'RIO_CAP_CLEAR 1\n|RIO_CAP_CLEAR 0\n'),
    ('f14b_mono', LcdButton.TWO, 'RIO_CAP_SW 1\n|RIO_CAP_SW 0\n'),
    ('f14b_mono', LcdButton.THREE, 'RIO_CAP_NE 1\n|RIO_CAP_NE 0\n'),
    ('f14b_mono', LcdButton.FOUR, 'RIO_CAP_ENTER 1\n|RIO_CAP_ENTER 0\n'),
    ('f14b_color', LcdButton.NONE, '\n'),
    ('f14b_color', LcdButton.LEFT, 'RIO_CAP_CLEAR 1\n|RIO_CAP_CLEAR 0\n'),
    ('f14b_color', LcdButton.RIGHT, 'RIO_CAP_SW 1\n|RIO_CAP_SW 0\n'),
    ('f14b_color', LcdButton.DOWN, 'RIO_CAP_NE 1\n|RIO_CAP_NE 0\n'),
    ('f14b_color', LcdButton.UP, 'RIO_CAP_ENTER 1\n|RIO_CAP_ENTER 0\n'),
    ('f14b_color', LcdButton.MENU, '\n'),
    ('f14b_color', LcdButton.CANCEL, '\n'),
    ('f14b_color', LcdButton.OK, '\n'),
    ('f16c50_mono', LcdButton.NONE, '\n'),
    ('f16c50_mono', LcdButton.ONE, 'IFF_MASTER_KNB 1\n'),
    ('f16c50_mono', LcdButton.TWO, 'IFF_ENABLE_SW 1\n'),
    ('f16c50_mono', LcdButton.THREE, 'IFF_M4_CODE_SW 1\n'),
    ('f16c50_mono', LcdButton.FOUR, 'IFF_M4_REPLY_SW 1\n'),
    ('f16c50_color', LcdButton.NONE, '\n'),
    ('f16c50_color', LcdButton.LEFT, 'IFF_MASTER_KNB 1\n'),
    ('f16c50_color', LcdButton.RIGHT, 'IFF_ENABLE_SW 1\n'),
    ('f16c50_color', LcdButton.DOWN, 'IFF_M4_CODE_SW 1\n'),
    ('f16c50_color', LcdButton.UP, 'IFF_M4_REPLY_SW 1\n'),
    ('f16c50_color', LcdButton.MENU, '\n'),
    ('f16c50_color', LcdButton.CANCEL, '\n'),
    ('f16c50_color', LcdButton.OK, '\n'),
    ('ah64dblkii_mono', LcdButton.NONE, '\n'),
    ('ah64dblkii_mono', LcdButton.ONE, 'PLT_EUFD_IDM 0\n|PLT_EUFD_IDM 1\n'),
    ('ah64dblkii_mono', LcdButton.TWO, 'PLT_EUFD_RTS 0\n|PLT_EUFD_RTS 1\n'),
    ('ah64dblkii_mono', LcdButton.THREE, 'PLT_EUFD_PRESET 0\n|PLT_EUFD_PRESET 1\n'),
    ('ah64dblkii_mono', LcdButton.FOUR, 'PLT_EUFD_ENT 0\n|PLT_EUFD_ENT 1\n'),
])
def test_button_pressed_for_planes(plane, button, result, request):
    plane = request.getfixturevalue(plane)
    assert plane.button_request(button) == result


@mark.parametrize('button, result', [
    (LcdButton.NONE, '\n'),
    (LcdButton.LEFT, 'PLT_EUFD_WCA 0\n|PLT_EUFD_WCA 1\n'),
    (LcdButton.RIGHT, 'PLT_EUFD_RTS 0\n|PLT_EUFD_RTS 1\n'),
    (LcdButton.DOWN, 'PLT_EUFD_PRESET 0\n|PLT_EUFD_PRESET 1\n'),
    (LcdButton.UP, 'PLT_EUFD_ENT 0\n|PLT_EUFD_ENT 1\n'),
    (LcdButton.MENU, '\n'),
    (LcdButton.CANCEL, '\n'),
    (LcdButton.OK, '\n'),
], ids=['NONE', 'LEFT', 'RIGHT', 'DOWN', 'UP', 'MENU', 'CANCEL', 'OK'])
def test_button_pressed_for_apache_color(button, result, ah64dblkii_color):
    from dcspy.aircraft import ApacheEufdMode
    ah64dblkii_color.mode = ApacheEufdMode.WCA
    assert ah64dblkii_color.button_request(button) == result


@mark.parametrize('plane, btn_name, btn, values', [
    ('f16c50_mono', 'IFF_MASTER_KNB', LcdButton.ONE, (1, 2, 3, 4, 3, 2, 1, 0, 1)),
    ('f16c50_mono', 'IFF_ENABLE_SW', LcdButton.TWO, (1, 2, 1, 0, 1)),
    ('f16c50_mono', 'IFF_M4_CODE_SW', LcdButton.THREE, (1, 2, 1, 0, 1)),
    ('f16c50_mono', 'IFF_M4_REPLY_SW', LcdButton.FOUR, (1, 2, 1, 0, 1)),
    ('f16c50_color', 'IFF_MASTER_KNB', LcdButton.LEFT, (1, 2, 3, 4, 3, 2, 1, 0, 1)),
    ('f16c50_color', 'IFF_ENABLE_SW', LcdButton.RIGHT, (1, 2, 1, 0, 1)),
    ('f16c50_color', 'IFF_M4_CODE_SW', LcdButton.DOWN, (1, 2, 1, 0, 1)),
    ('f16c50_color', 'IFF_M4_REPLY_SW', LcdButton.UP, (1, 2, 1, 0, 1)),
    ('fa18chornet_color', 'HUD_ATT_SW', LcdButton.OK, (1, 2, 1, 0, 1)),
    ('fa18chornet_color', 'IFEI_DWN_BTN', LcdButton.MENU, (1, 0, 1)),
    ('fa18chornet_color', 'IFEI_UP_BTN', LcdButton.CANCEL, (1, 0, 1)),
], ids=[
    'ONE - Viper Mono',
    'TWO - Viper Mono',
    'THREE - Viper Mono',
    'FOUR - Viper Mono',
    'LEFT - Viper Color',
    'RIGHT - Viper Color',
    'DOWN - Viper Color',
    'UP - Viper Color',
    'OK - Hornet Color',
    'MENU - Hornet Color',
    'CANCEL - Hornet Color',
])
def test_get_next_value_for_cycle_buttons(plane, btn_name, btn, values, request):
    plane = request.getfixturevalue(plane)
    assert not all(isinstance(cyc_btn.iter, ZigZagIterator) for cyc_btn in plane.cycle_buttons.values())
    for val in values:
        assert plane.button_request(btn) == f'{btn_name} {val}\n'
    assert isinstance(plane.cycle_buttons[btn].iter, ZigZagIterator)


# <=><=><=><=><=> Set BIOS <=><=><=><=><=>
@mark.parametrize('plane, bios_pairs, result', [
    ('fa18chornet_mono', [('UFC_SCRATCHPAD_STRING_2_DISPLAY', '~~')], '22'),
    ('fa18chornet_mono', [('UFC_COMM1_DISPLAY', '``')], '11'),
    ('fa18chornet_mono', [('IFEI_FUEL_UP', '104T')], '104T'),
    ('fa18chornet_color', [('UFC_SCRATCHPAD_STRING_1_DISPLAY', '~~')], '22'),
    ('fa18chornet_color', [('UFC_COMM1_DISPLAY', '``')], '11'),
    ('fa18chornet_color', [('IFEI_FUEL_UP', '1000T')], '1000T'),
    ('f16c50_mono', [('DED_LINE_1', 'a')], '\u2666'),
    ('f16c50_mono', [('DED_LINE_2', 'o')], '\u00b0'),
    ('f16c50_mono', [('DED_LINE_3', '*')], '\u25d9'),
    ('f16c50_mono', [('DED_LINE_5', '\x07')], ''),
    ('f16c50_mono', [('DED_LINE_4', '\x10')], ''),
    ('f16c50_mono', [('DED_LINE_2', '   @')], '   '),
    ('f16c50_color', [('DED_LINE_3', 'a')], '\u0040'),
    ('f16c50_color', [('DED_LINE_4', 'o')], '\u005e'),
    ('f16c50_color', [('DED_LINE_3', '*')], '\u00d7'),
    ('f16c50_color', [('DED_LINE_5', '\xfe')], ''),
    ('f16c50_color', [('DED_LINE_1', '\xfc')], ''),
    ('f16c50_color', [('DED_LINE_2', '   @')], '   '),
    ('f16c50_color', [('DED_LINE_2', '1DEST 2BNGO 3VIP  RINTG  A\x10\x04')], 'ÁDEST ÂBNGO ÃVIP  rINTG  '),
    ('f16c50_color', [('DED_LINE_3', '4NAV  5MAN  6INS  EDLNK  A\x10\x04')], 'ÄNAV  ÅMAN  ÆINS  eDLNK  '),
    ('f16c50_color', [('DED_LINE_4', '7CMDS 8MODE 9VRP  0MISC  A\x10\x04')], 'ÇCMDS ÈMODE ÉVRP  ÀMISC  '),
    ('f16c50_mono', [('DED_LINE_1', '       *      *CMD STRG  \x80@')], '       \u25d9      \u25d9CMD STRG  '),
    ('f16c50_mono', [('DED_LINE_2', '1DEST 2BNGO 3VIP  RINTG  A\x10\x04')], '1DEST 2BNGO 3VIP  RINTG  '),
    ('f16c50_mono', [('DED_LINE_1', ' MARK *HUD *    26a      @')], ' MARK \u25d9HUD \u25d9    26\u2666      '),
    ('f16c50_mono', [('DED_LINE_5', 'M3 :7000 *     *DCPL(9)  \x03\x82')], 'M3 :7000 \u25d9     \u25d9DCPL(9)  '),
    ('ah64dblkii_mono', [('PLT_EUFD_LINE8', '~=>VHF*  121.000   -----              121.500   -----   ')], '\u25a0\u2219\u25b8VHF*  121.000   -----              121.500   -----   '),
    ('ah64dblkii_mono', [('PLT_EUFD_LINE9', ' <=UHF*  305.000   -----              305.000   -----   ')], ' \u25c2\u2219UHF*  305.000   -----              305.000   -----   '),
    ('ah64dblkii_mono', [('PLT_EUFD_LINE10', ' <>FM1*   30.000   -----    NORM       30.000   -----   ')], ' \u25c2\u25b8FM1*   30.000   -----    NORM       30.000   -----   '),
    ('ah64dblkii_color', [('PLT_EUFD_LINE11', '[==FM2*   30.000   -----               30.000   -----   ')], '\u25ca\u2219\u2219FM2*   30.000   -----               30.000   -----   '),
    ('ah64dblkii_color', [('PLT_EUFD_LINE12', ' ==HF *    2.0000A -----    LOW         2.0000A -----   ')], ' \u2219\u2219HF *    2.0000A -----    LOW         2.0000A -----   '),
    ('ah64dblkii_color', [('PLT_EUFD_LINE12', ']==HF *    2.0000A -----    LOW         2.0000A -----   ')], '\u2666\u2219\u2219HF *    2.0000A -----    LOW         2.0000A -----   '),
])
def test_set_bios_for_airplane(plane, bios_pairs, result, request):
    plane = request.getfixturevalue(plane)
    set_bios_during_test(plane, bios_pairs)
    assert plane.bios_data[bios_pairs[0][0]] == result


@mark.parametrize('plane, bios_pairs, mode', [
    ('ah64dblkii_mono', [('PLT_EUFD_LINE1', 'ENGINE 1 OUT      |AFT FUEL LOW      |TAIL WHL LOCK SEL ')], 'IDM'),
    ('ah64dblkii_mono', [('PLT_EUFD_LINE1', '                  |AFT FUEL LOW      |PRESET TUNE VHS ')], 'PRE'),
    ('ah64dblkii_color', [('PLT_EUFD_LINE1', '                  |                  |TAIL WHL LOCK SEL ')], 'IDM'),
    ('ah64dblkii_color', [('PLT_EUFD_LINE1', '                  |AFT FUEL LOW      |TAIL WHL LOCK SEL ')], 'IDM'),
    ('ah64dblkii_color', [('PLT_EUFD_LINE1', 'ENGINE 1 OUT      |AFT FUEL LOW      |PRESET TUNE FM1 ')], 'PRE'),
], ids=['Mono IDM', 'Mono PRE', 'Color IDM 1', 'Color IDM 2', 'Color PRE'])
def test_apache_mode_switch_idm_pre_for_apache(plane, bios_pairs, mode, request):
    plane = request.getfixturevalue(plane)
    set_bios_during_test(plane, bios_pairs)
    assert plane.mode.name == mode


# <=><=><=><=><=> Prepare Image <=><=><=><=><=>

@mark.parametrize('lcd', ['mono', 'color'])
@mark.parametrize('model', all_plane_list)
def test_prepare_image_for_all_planes(model, lcd, resources, img_precision, request):
    aircraft_model = request.getfixturevalue(f'{model}_{lcd}')
    bios_pairs = request.getfixturevalue(f'{model}_{lcd}_bios')
    set_bios_during_test(aircraft_model, bios_pairs)
    img = aircraft_model.prepare_image()
    # if 'f14b' in model or 'f14a135gr' in model:
    #     img.save(resources / platform / f'{model}_{lcd}_{type(aircraft_model).__name__}.png')
    # else:
    assert compare_images(img=img, file_path=resources / platform / f'{model}_{lcd}_{type(aircraft_model).__name__}.png', precision=img_precision)


@mark.parametrize('model', ['ah64dblkii_mono', 'ah64dblkii_color'], ids=['Mono LCD', 'Color LCD'])
def test_prepare_image_for_apache_wca_mode(model, resources, img_precision, request):
    from itertools import repeat
    from tempfile import gettempdir

    from dcspy.aircraft import ApacheEufdMode

    apache = request.getfixturevalue(model)
    apache._debug_img = repeat(999)
    bios_pairs = [
        ('PLT_EUFD_LINE1', 'LOW ROTOR RPM     |RECTIFIER 2 FAIL  |CHARGER           '),
        ('PLT_EUFD_LINE2', 'ENGINE 2 OUT      |GENERATOR 2 FAIL  |TAIL WHL LOCK SEL '),
        ('PLT_EUFD_LINE3', 'ENGINE 1 OUT      |AFT FUEL LOW      |                  '),
        ('PLT_EUFD_LINE4', '                  |FORWARD FUEL LOW  |                  '),
        ('PLT_EUFD_LINE5', '                  |                  |                  '),
    ]
    set_bios_during_test(apache, bios_pairs)
    apache.mode = ApacheEufdMode.WCA
    apache.cfg['save_lcd'] = True
    img = apache.prepare_image()
    assert (Path(gettempdir()) / f'{type(apache).__name__}_999.png').exists()
    assert compare_images(img=img, file_path=resources / platform / f'{model}_wca_mode.png', precision=img_precision)


# <=><=><=><=><=> Apache special <=><=><=><=><=>
@mark.parametrize('model', ['ah64dblkii_mono', 'ah64dblkii_color'], ids=['Mono LCD', 'Color LCD'])
def test_apache_wca_more_then_one_screen_scrolled(model, resources, img_precision, request):
    from dcspy.aircraft import ApacheEufdMode
    apache = request.getfixturevalue(model)
    bios_pairs = [
        ('PLT_EUFD_LINE1', 'LOW ROTOR RPM     |RECTIFIER 2 FAIL  |CHARGER           '),
        ('PLT_EUFD_LINE2', 'ENGINE 2 OUT      |GENERATOR 2 FAIL  |TAIL WHL LOCK SEL '),
        ('PLT_EUFD_LINE3', 'ENGINE 1 OUT      |AFT FUEL LOW      |                  '),
    ]
    set_bios_during_test(apache, bios_pairs)
    apache.mode = ApacheEufdMode.WCA

    for i in range(1, 3):
        assert apache.warning_line == i
        apache.warning_line += 1
        apache.prepare_image()
    assert apache.warning_line == 3
    img = apache.prepare_image()
    assert compare_images(img=img, file_path=resources / platform / f'{model}_wca_mode_scroll.png', precision=img_precision)


@mark.parametrize('model', ['ah64dblkii_mono', 'ah64dblkii_color'], ids=['Mono LCD', 'Color LCD'])
def test_apache_pre_mode(model, apache_pre_mode_bios_data, resources, img_precision, request):
    apache = request.getfixturevalue(model)
    set_bios_during_test(apache, apache_pre_mode_bios_data)
    img = apache.prepare_image()
    assert compare_images(img=img, file_path=resources / platform / f'{model}_pre_mode.png', precision=img_precision)



from os import linesep, makedirs
from pathlib import Path
from unittest.mock import MagicMock, PropertyMock, mock_open, patch

import pytest
from packaging import version
from pytest import mark

from dcspy import utils
from dcspy.models import DEFAULT_FONT_NAME, ReleaseInfo


@mark.parametrize('online_tag, result', [
    ('1.1.1', ReleaseInfo(
        latest=True, ver=version.parse('1.1.1'), dl_url='github.com/fake.tgz', published='09 August 2021', release_type='Pre-release', asset_file='fake.tgz'
    )),
    ('3.2.1', ReleaseInfo(
        latest=False, ver=version.parse('3.2.1'), dl_url='github.com/fake.tgz', published='09 August 2021', release_type='Pre-release', asset_file='fake.tgz'
    ))
], ids=[
    'No update',
    'New version',
])
def test_check_ver_is_possible(online_tag, result):
    with patch.object(utils, 'get') as response_get:
        type(response_get.return_value).ok = PropertyMock(return_value=True)
        type(response_get.return_value).json = MagicMock(return_value={'tag_name': online_tag, 'prerelease': True,
                                                                       'assets': [{
                                                                           'browser_download_url': 'github.com/fake.tgz'}],
                                                                       'published_at': '2021-08-09T16:41:51Z'})
        assert utils.check_ver_at_github(repo='fake1/package1', current_ver='1.1.1', extension='.tgz') == result


def test_check_ver_can_not_check():
    with patch.object(utils, 'get') as response_get:
        type(response_get.return_value).ok = PropertyMock(return_value=False)
        rel_info = utils.check_ver_at_github(repo='fake2/package2', current_ver='2.2.2', extension='.zip')
        assert rel_info == ReleaseInfo(latest=False, ver=version.parse('0.0.0'), dl_url='', published='', release_type='Regular', asset_file='')


def test_check_ver_exception():
    with patch.object(utils, 'get', side_effect=Exception('Connection error')):
        rel_info = utils.check_ver_at_github(repo='fake3/package3', current_ver='3.3.3', extension='.exe')
        assert rel_info == ReleaseInfo(latest=False, ver=version.parse('0.0.0'), dl_url='', published='', release_type='Regular', asset_file='')


@mark.parametrize('online_tag, result', [
    ('1.1.1', 'v1.1.1 (latest)'),
    ('3.2.1', 'v1.1.1 (update!)')
], ids=['No update', 'New version'])
def test_get_version_string_is_possible(online_tag, result):
    with patch.object(utils, 'get') as response_get:
        type(response_get.return_value).ok = PropertyMock(return_value=True)
        type(response_get.return_value).json = MagicMock(return_value={'tag_name': online_tag, 'prerelease': True,
                                                                       'assets': [{
                                                                           'browser_download_url': 'github.com/fake.tgz'}],
                                                                       'published_at': '2021-08-09T16:41:51Z'})
        assert utils.get_version_string(repo='fake1/package1', current_ver='1.1.1', check=True) == result


def test_get_version_string_without_checking():
    assert utils.get_version_string(repo='fake4/package4', current_ver='4.4.4', check=False) == 'v4.4.4'


def test_get_version_string_exception():
    with patch.object(utils, 'get', side_effect=Exception('Connection error')):
        assert utils.get_version_string(repo='fake4/package4', current_ver='4.4.4', check=True) == 'v4.4.4 (failed)'


@mark.parametrize('response, result', [(False, False), (True, True)], ids=['Download failed', 'Download success'])
def test_download_file(response, result, tmp_path):
    dl_file = tmp_path / 'tmp.txt'
    with patch.object(utils, 'get') as response_get:
        type(response_get.return_value).ok = PropertyMock(return_value=response)
        type(response_get.return_value).iter_content = MagicMock(return_value=[b'1', b'0', b'0', b'1'])
        assert utils.download_file('https://test.com', dl_file) is result


def test_proc_is_running():
    assert utils.proc_is_running('python')
    assert not utils.proc_is_running('wrong_python')


def test_dummy_save_load_migrate(tmpdir):
    from os import environ

    from dcspy.migration import migrate
    test_tmp_yaml = Path(tmpdir) / 'test_cfg.yaml'

    utils.save_yaml(data={'font_mono_s': 9}, full_path=test_tmp_yaml)
    d_cfg = utils.load_yaml(full_path=test_tmp_yaml)
    assert d_cfg == {'font_mono_s': 9}
    d_cfg = migrate(cfg=d_cfg)
    assert d_cfg == {
        'api_ver': '3.0.0',
        'keyboard': 'G13',
        'save_lcd': False,
        'show_gui': True,
        'autostart': False,
        'completer_items': 20,
        'current_plane': 'A-10C',
        'dcsbios': f'D:\\Users\\{environ.get("USERNAME", "UNKNOWN")}\\Saved Games\\DCS.openbeta\\Scripts\\DCS-BIOS',
        'dcs': 'C:/Program Files/Eagle Dynamics/DCS World OpenBeta',
        'verbose': False,
        'check_bios': True,
        'check_ver': True,
        'font_name': DEFAULT_FONT_NAME,
        'font_mono_m': 9,
        'font_mono_s': 9,
        'font_mono_l': 16,
        'font_color_m': 22,
        'font_color_s': 18,
        'font_color_l': 32,
        'f16_ded_font': True,
        'git_bios': True,
        'toolbar_area': 4,
        'toolbar_style': 0,
        'git_bios_ref': 'master',
        'gkeys_area': 2,
        'gkeys_float': False,
    }
    with open(test_tmp_yaml, 'w+') as f:
        f.write('')
    d_cfg = utils.load_yaml(full_path=test_tmp_yaml)
    assert len(d_cfg) == 0


def test_check_dcs_ver_file_exists_with_ver(autoupdate1_cfg):
    with patch('dcspy.utils.open', mock_open(read_data=autoupdate1_cfg)):
        dcs_ver = utils.check_dcs_ver(Path(''))
        assert dcs_ver == ('openbeta', '2.7.16.28157')


def test_check_dcs_ver_file_exists_without_ver(autoupdate2_cfg):
    with patch('dcspy.utils.open', mock_open(read_data=autoupdate2_cfg)):
        dcs_ver = utils.check_dcs_ver(Path(''))
        assert dcs_ver == ('openbeta', 'Unknown')


def test_check_dcs_ver_file_exists_without_branch(autoupdate3_cfg):
    with patch('dcspy.utils.open', mock_open(read_data=autoupdate3_cfg)):
        dcs_ver = utils.check_dcs_ver(Path(''))
        assert dcs_ver == ('stable', '2.7.18.28157')


@mark.parametrize('side_effect', [FileNotFoundError, PermissionError])
def test_check_dcs_ver_file_not_exists(side_effect):
    with patch('dcspy.utils.open', side_effect=side_effect):
        dcs_ver = utils.check_dcs_ver(Path(''))
        assert dcs_ver == ('Unknown', 'Unknown')


def test_check_bios_ver(tmpdir):
    makedirs(Path(tmpdir) / 'lib' / 'modules' / 'common_modules')
    common_data_lua = Path(tmpdir) / 'lib' / 'modules' / 'common_modules' / 'CommonData.lua'
    with open(file=common_data_lua, encoding='utf-8', mode='w+') as cd_lua:
        cd_lua.write('local function getVersion()\n\treturn "1.2.3"\nend')
    result = utils.check_bios_ver(bios_path=tmpdir)
    assert result == ReleaseInfo(latest=False, ver=version.parse('1.2.3'), dl_url='', published='', release_type='', asset_file='')


def test_check_bios_ver_empty_lua(tmpdir):
    makedirs(Path(tmpdir) / 'lib')
    common_data_lua = Path(tmpdir) / 'lib' / 'CommonData.lua'
    with open(file=common_data_lua, encoding='utf-8', mode='w+') as cd_lua:
        cd_lua.write('')
    result = utils.check_bios_ver(bios_path=tmpdir)
    assert result == ReleaseInfo(latest=False, ver=version.parse('0.0.0'), dl_url='', published='', release_type='', asset_file='')


def test_check_bios_ver_raise_exception(tmpdir):
    result = utils.check_bios_ver(bios_path=tmpdir)
    assert result == ReleaseInfo(latest=False, ver=version.parse('0.0.0'), dl_url='', published='', release_type='', asset_file='')


def test_is_git_repo(tmpdir):
    from git import Repo
    assert utils.is_git_repo(tmpdir) is False
    assert utils.is_git_repo(tmpdir / 'new') is False
    Repo.init(tmpdir)
    assert utils.is_git_repo(tmpdir) is True


def test_is_git_exec_present():
    assert utils.is_git_exec_present() is True


@mark.slow
def test_check_github_repo(tmpdir):
    from re import search
    sha = utils.check_github_repo(git_ref='master', update=True, repo='emcek/common_sense', repo_dir=tmpdir)
    match = search(r'(master):\s\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}\+\d{2}:\d{2}\sby:\s.*', sha)
    assert match.group(1) == 'master'
    sha = utils.check_github_repo(git_ref='branch', update=True, repo='emcek/common_sense', repo_dir=tmpdir)
    match = search(r'([0-9a-f]{8})\sfrom:\s\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}\+\d{2}:\d{2}\sby:\s.*', sha)
    assert match.group(1)
    sha = utils.check_github_repo(git_ref='master', update=False, repo='emcek/common_sense', repo_dir=tmpdir)
    match = search(r'([0-9a-f]{8})\sfrom:\s\d{2}-\w{3}-\d{4}\s\d{2}:\d{2}:\d{2}', sha)
    assert match.group(1)


@mark.slow
def test_is_git_object(tmpdir):
    utils.check_github_repo(git_ref='master', update=True, repo='emcek/common_sense', repo_dir=tmpdir)
    assert utils.is_git_object(repo_dir=tmpdir, git_obj='master') is True
    assert utils.is_git_object(repo_dir=tmpdir, git_obj='wrong') is False
    assert utils.is_git_object(repo_dir=Path('/'), git_obj='master') is False


@mark.slow
def test_get_all_git_refs(tmpdir):
    utils.check_github_repo(git_ref='master', update=True, repo='emcek/common_sense', repo_dir=tmpdir)
    assert utils.get_all_git_refs(repo_dir=tmpdir) == ['master']


@mark.slow
def test_get_sha_for_current_git_ref(tmpdir):
    hex_sha = utils.get_sha_for_current_git_ref(git_ref='master', repo='emcek/common_sense', repo_dir=tmpdir)
    assert len(hex_sha) == 40
    assert hex_sha == '6d5d9f80a8309eec3e9cf183fc57d56559035560'


def test_check_dcs_bios_entry_no_entry(tmpdir):
    from os import makedirs
    install_dir = tmpdir / 'install'
    makedirs(install_dir)
    lua = 'Export.lua'
    lua_dst_data = ''

    with open(file=tmpdir / lua, mode='a+', encoding='utf-8') as lua_from_zip:
        lua_from_zip.write('anything')
    with open(file=install_dir / lua, mode='a+', encoding='utf-8') as lua_dst:
        lua_dst.write(lua_dst_data)

    result = utils.check_dcs_bios_entry(lua_dst_data=lua_dst_data, lua_dst_path=install_dir, temp_dir=tmpdir)
    assert result == '\n\nExport.lua exists.\n\nDCS-BIOS entry added.\n\nYou verify installation at:\ngithub.com/DCS-Skunkworks/DCSFlightpanels/wiki/Installation'


def test_check_dcs_bios_entry_ok(tmpdir):
    from os import makedirs
    install_dir = tmpdir / 'install'
    makedirs(install_dir)
    lua = 'Export.lua'
    lua_dst_data = r'dofile(lfs.writedir()..[[Scripts\DCS-BIOS\BIOS.lua]])'

    with open(file=tmpdir / lua, mode='a+', encoding='utf-8') as lua_from_zip:
        lua_from_zip.write('anything')
    with open(file=install_dir / lua, mode='a+', encoding='utf-8') as lua_dst:
        lua_dst.write(lua_dst_data)

    result = utils.check_dcs_bios_entry(lua_dst_data=lua_dst_data, lua_dst_path=install_dir, temp_dir=tmpdir)
    assert result == '\n\nExport.lua exists.\n\nDCS-BIOS entry detected.'


@mark.slow
def test_collect_debug_data():
    from tempfile import gettempdir
    from zipfile import ZipFile
    with open(Path(gettempdir()) / 'Ka50_999.png', 'w+') as png:
        png.write('')
    zip_file = utils.collect_debug_data()
    assert 'dcspy_debug_' in str(zip_file)
    assert zip_file.suffix == '.zip'
    assert zip_file.is_file()
    assert zip_file.exists()
    with ZipFile(file=zip_file, mode='r') as zipf:
        zip_list = zipf.namelist()
    assert 'system_data.txt' in zip_list
    assert 'config.yaml' in zip_list
    assert 'dcspy.log' in zip_list
    assert 'Ka50_999.png' in zip_list


@mark.slow
def test_run_pip_command_success():
    rc, err, out = utils.run_pip_command('list')
    assert rc == 0
    assert 'pip' in out, out
    assert err == '' or err == f'WARNING: There was an error checking the latest version of pip.{linesep}', err


@mark.slow
def test_run_pip_command_failed():
    rc, err, out = utils.run_pip_command('bullshit')
    assert rc == 1
    assert out == '', out
    assert err != '', err


def test_get_full_bios_for_plane(resources):
    a10_model = utils.get_full_bios_for_plane(plane='A-10C', bios_dir=resources / 'dcs_bios')
    assert len(a10_model.root) == 64
    assert sum(len(values) for values in a10_model.root.values()) == 772


def test_get_inputs_for_plane(resources):
    from dcspy.models import CTRL_LIST_SEPARATOR, ControlKeyData
    bios = utils.get_inputs_for_plane(plane='A-10C', bios_dir=resources / 'dcs_bios')
    for section, ctrls in bios.items():
        for name, ctrl in ctrls.items():
            assert isinstance(ctrl, ControlKeyData), f'Wrong type fpr {section} / {name}'

    list_of_ctrls = utils.get_list_of_ctrls(inputs=bios)
    assert CTRL_LIST_SEPARATOR
    assert CTRL_LIST_SEPARATOR in list_of_ctrls[0], list_of_ctrls[0]


def test_get_inputs_for_wrong_plane(resources):
    with pytest.raises(KeyError):
        _ = utils.get_inputs_for_plane(plane='Wrong', bios_dir=resources / 'dcs_bios')


def test_get_plane_aliases_all(resources):
    s = utils.get_plane_aliases(bios_dir=resources / 'dcs_bios')
    assert s == {
        'A-10C': ['CommonData', 'A-10C'],
        'A-10C_2': ['CommonData', 'A-10C'],
        'AH-64D_BLK_II': ['CommonData', 'AH-64D'],
        'AV8BNA': ['CommonData', 'AV8BNA', 'NS430'],
        'F-14A-135-GR': ['CommonData', 'F-14', 'NS430'],
        'F-14B': ['CommonData', 'F-14', 'NS430'],
        'F-15ESE': ['CommonData', 'F-15E'],
        'F-16C_50': ['CommonData', 'F-16C_50'],
        'FA-18C_hornet': ['CommonData', 'FA-18C_hornet'],
        'Ka-50': ['CommonData', 'Ka-50'],
        'Ka-50_3': ['CommonData', 'Ka-50'],
        'Mi-24P': ['CommonData', 'Mi-24P', 'NS430'],
        'Mi-8MT': ['CommonData', 'Mi-8MT', 'NS430'],
        'NS430': ['CommonData'],
    }


def test_get_plane_aliases_one_plane(resources):
    s = utils.get_plane_aliases(bios_dir=resources / 'dcs_bios', plane='A-10C')
    assert s == {'A-10C': ['CommonData', 'A-10C']}


def test_get_plane_aliases_wrong_plane(resources):
    with pytest.raises(KeyError):
        _ = utils.get_plane_aliases(bios_dir=resources / 'dcs_bios', plane='A-Wrong')


def test_get_planes_list(resources):
    plane_list = utils.get_planes_list(bios_dir=resources / 'dcs_bios')
    assert plane_list == [
        'A-10C',
        'A-10C_2',
        'AH-64D_BLK_II',
        'AV8BNA',
        'F-14A-135-GR',
        'F-14B',
        'F-15ESE',
        'F-16C_50',
        'FA-18C_hornet',
        'Ka-50',
        'Ka-50_3',
        'Mi-24P',
        'Mi-8MT',
    ]



from sys import platform
from unittest.mock import patch

from pytest import mark


@mark.slow
@mark.qt6
@mark.skipif(condition=platform != 'win32', reason='Run only on Windows')
def test_qt(qtbot, test_config_yaml, switch_dcs_bios_path_in_config, tmpdir):
    from time import sleep

    from PySide6 import QtCore
    from PySide6.QtCore import Qt
    from PySide6.QtQuick import QQuickWindow, QSGRendererInterface

    from dcspy import qt_gui

    QQuickWindow.setGraphicsApi(QSGRendererInterface.OpenGLRhi)
    QtCore.QCoreApplication.setAttribute(Qt.AA_ShareOpenGLContexts)

    with patch('dcspy.qt_gui.default_yaml', test_config_yaml):
        with patch.object(qt_gui.DcsPyQtGui, '_run_file_dialog', return_value=tmpdir):
            dcspy_gui = qt_gui.DcsPyQtGui()
            dcspy_gui.show()
            qtbot.addWidget(dcspy_gui)
            dcspy_gui.rb_g19.setChecked(True)
            dcspy_gui.sp_completer.setValue(10)
            dcspy_gui.combo_planes.setCurrentIndex(1)
            sleep(0.7)
            qtbot.mouseClick(dcspy_gui.pb_start, Qt.LeftButton)
            qtbot.mouseClick(dcspy_gui.pb_stop, Qt.LeftButton)
            qtbot.mouseClick(dcspy_gui.pb_collect_data, Qt.LeftButton)
            dcspy_gui.dw_gkeys.show()

            dcspy_gui.tw_gkeys.cellWidget(0, 0).setCurrentText('ADI_PITCH_TRIM')
            qtbot.mouseClick(dcspy_gui.tw_gkeys.cellWidget(0, 0), Qt.LeftButton)
            qtbot.mouseClick(dcspy_gui.rb_set_state, Qt.LeftButton)

            dcspy_gui.tw_gkeys.cellWidget(0, 1).setCurrentText('ARC210_CHN_KNB')

            dcspy_gui.tw_gkeys.cellWidget(1, 0).setCurrentIndex(5)
            qtbot.mouseClick(dcspy_gui.tw_gkeys.cellWidget(1, 0), Qt.LeftButton)
            qtbot.mouseClick(dcspy_gui.rb_fixed_step_dec, Qt.LeftButton)
            dcspy_gui.current_row = 1
            dcspy_gui.current_col = 0
            qtbot.mouseClick(dcspy_gui.pb_copy, Qt.LeftButton)
            qtbot.mouseClick(dcspy_gui.pb_save, Qt.LeftButton)

            qtbot.mouseClick(dcspy_gui.pb_close, Qt.LeftButton)
            sleep(0.7)



import socket
from unittest.mock import patch


def test_load_new_plane_if_detected():
    from dcspy import starter
    with patch.object(starter, 'KeyboardManager') as lcd:
        starter._load_new_plane_if_detected(lcd)
        lcd.load_new_plane.assert_called_once_with()


def test_sock_err_handler(keyboard_mono):
    from time import time

    from dcspy import starter
    ver_string = f'v{starter.__version__} (latest)'
    start_time = time()
    starter._sock_err_handler(manager=keyboard_mono, start_time=start_time, ver_string=ver_string,
                              support_iter=(i for i in '12'), exp=Exception())
    assert keyboard_mono.display == ['Logitech LCD OK', 'No data from DCS:   00:00', '1', ver_string]


def test_supporters():
    from dcspy import starter
    sup_iter = starter._supporters(text='123456', width=5)
    assert next(sup_iter) == '12345'
    assert next(sup_iter) == '23456'
    assert next(sup_iter) == '34561'


def test_prepare_socket():
    from dcspy import starter
    sock = starter._prepare_socket()
    assert isinstance(sock, socket.socket)
    assert sock.proto == 17
    assert sock.type in (2050, 2)
    assert sock.family == 2



from ctypes import c_void_p, sizeof
from dataclasses import dataclass
from logging import getLogger
from os import environ
from pathlib import Path
from platform import architecture
from sys import maxsize
from typing import Optional

from _cffi_backend import Lib
from cffi import FFI

LOG = getLogger(__name__)


@dataclass
class DllSdk:
    """DLL SDK."""
    name: str
    header: str
    dir: str


with open(file=Path(__file__).resolve().with_name('LogitechLCDLib.h')) as lcd_header_file:
    lcd_header = lcd_header_file.read()
with open(file=Path(__file__).resolve().with_name('LogitechLEDLib.h')) as led_header_file:
    led_header = led_header_file.read()
with open(file=Path(__file__).resolve().with_name('LogitechGkeyLib.h')) as led_header_file:
    key_header = led_header_file.read()

LcdDll = DllSdk(name='LCD', dir='LCD', header=lcd_header)
LedDll = DllSdk(name='LED', dir='LED', header=led_header)
KeyDll = DllSdk(name='Gkey', dir='G-key', header=key_header)


def load_dll(lib_type: DllSdk) -> Optional[Lib]:
    """
    Initialize and load of C dynamic linking library.

    :param lib_type: library to load: LCD or LED
    :return: C DLL instance
    """
    try:
        arch = 'x64' if all([architecture()[0] == '64bit', maxsize > 2 ** 32, sizeof(c_void_p) > 4]) else 'x86'
        try:
            prog_files = environ['PROGRAMW6432']
        except KeyError:
            prog_files = environ['PROGRAMFILES']
        dll_path = f'{prog_files}\\Logitech Gaming Software\\SDK\\{lib_type.dir}\\{arch}\\Logitech{lib_type.name.capitalize()}.dll'
        LOG.debug(f'Selected DLL: {dll_path}')
        ffi = FFI()
        ffi.cdef(lib_type.header)
        dll = ffi.dlopen(dll_path)
        LOG.info(f'Loading of {lib_type.name} SDK success')
        return dll
    except (KeyError, OSError) as err:
        header = '*' * 44
        LOG.error(f'\n{header}\n*{type(err).__name__:^42}*\n{header}\nLoading of {lib_type.name} SDK failed !', exc_info=True)
        LOG.error(f'{header}')
        return None



from logging import getLogger
from pprint import pformat
from typing import Callable, Iterator, Union

from packaging import version

from dcspy.models import DcspyConfigYaml
from dcspy.utils import defaults_cfg

LOG = getLogger(__name__)
__version__ = '3.0.0'


def migrate(cfg: DcspyConfigYaml) -> DcspyConfigYaml:
    """
    Perform migration of configuration based on API version.

    If api_ver key do not exist, it is set to 2.3.3.

    :param cfg: configuration dict
    :return: Full migrated dict
    """
    LOG.debug(f'Starting configuration:\n{pformat(cfg)}')
    src_ver = cfg.get('api_ver', '2.3.3')  # do not touch this api_ver!
    LOG.debug(f'Current API version: {src_ver}')
    for migration_func in _filter_api_ver_func(str(src_ver)):
        migration_func(cfg)
        LOG.debug(f'Migration done: {migration_func.__name__}')
    cfg['api_ver'] = __version__
    cfg_with_defaults = {key: cfg.get(key, value) for key, value in defaults_cfg.items()}
    if 'UNKNOWN' in str(cfg_with_defaults['dcsbios']):
        cfg_with_defaults['dcsbios'] = defaults_cfg['dcsbios']
    final_cfg = {**cfg_with_defaults, **cfg}
    LOG.debug(f'Final configuration:\n{pformat(final_cfg)}')
    return final_cfg


def _filter_api_ver_func(cfg_ver: str) -> Iterator[Callable[[DcspyConfigYaml], None]]:
    """
    Filter migration function to call.

    :param cfg_ver: Current version of configuration
    :return: yield list of migration functions
    """
    api_ver_list = sorted([func_name.strip('_api_ver_').replace('_', '.')
                           for func_name in globals()
                           if func_name.startswith('_api_ver_')],
                          key=version.Version)
    for api_ver in api_ver_list:
        if version.Version(api_ver) > version.Version(cfg_ver) <= version.Version(__version__):
            yield globals()['_api_ver_{}'.format(api_ver.replace('.', '_'))]


def _api_ver_3_0_0rc1(cfg: DcspyConfigYaml) -> None:
    """
    Migrate to version 3.0.0.

    :param cfg: Configuration dictionary
    """
    _add_key(cfg, 'completer_items', 20)
    _add_key(cfg, 'current_plane', 'A-10C')
    _add_key(cfg, 'gkeys_area', 2)
    _add_key(cfg, 'gkeys_float', False)
    _add_key(cfg, 'toolbar_area', 4)
    _add_key(cfg, 'toolbar_style', 0)

    _remove_key(cfg, 'theme_color')
    _remove_key(cfg, 'theme_mode')

    _rename_key_keep_value(cfg, 'font_color_s', 'font_color_m', 22)
    _rename_key_keep_value(cfg, 'font_color_xs', 'font_color_s', 18)
    _rename_key_keep_value(cfg, 'font_mono_s', 'font_mono_m', 11)
    _rename_key_keep_value(cfg, 'font_mono_xs', 'font_mono_s', 9)


def _add_key(cfg: DcspyConfigYaml, key: str, default_value: Union[str, int, bool]) -> None:
    """
    Add key to dictionary if not exists.

    :param cfg: Configuration dictionary
    :param key: key name
    :param default_value: value to set
    """
    if key not in cfg:
        cfg[key] = default_value


def _remove_key(cfg: DcspyConfigYaml, key: str) -> None:
    """
    Remove key from dictionary.

    :param cfg: Configuration dictionary
    :param key: key name
    """
    try:
        del cfg[key]
    except KeyError:
        pass


def _rename_key_keep_value(cfg: DcspyConfigYaml, old_name: str, new_name: str, default_value: Union[str, int, bool]) -> None:
    """
    Rename key in dictionary and keep value.

    :param cfg: Configuration dictionary
    :param old_name: Old key name
    :param new_name: New key name
    :param default_value: default value if old key do not exist
    """
    value = cfg.get(old_name, default_value)
    try:
        del cfg['old_name']
    except KeyError:
        pass
    cfg[new_name] = value
